from . import lib
from .lib import (
    Sketcher,
    Profiler,
    Database,
    DatabaseFile,
    Sketch,
    GenomeSketch,
    SampleSketch,
    AniResult,
    ProfileResult,
)

__version__ = lib.__version__
__author__ = lib.__author__
__doc__ = lib.__doc__
# __build__ = lib.__build__
__all__ = [
    "Sketcher",
    "Profiler",
    "Database",
    "DatabaseFile",
    "Sketch",
    "GenomeSketch",
    "SampleSketch",
    "AniResult",
    "ProfileResult",
]
