//! This module contains code taken verbatim from the `sylph` source.
//!
//! The code here is re-used and redistributed under the terms of the MIT
//! license. All credits and copyright to Jim Shaw and Yun William Yu.

pub mod contain;
pub mod sketch;
